@@ -1,6 +1,6 @@
 import numpy as np
 
-from ...types import *
+from ...misc import *
 from ....types import *
 from ..base import Converter
 from ....misc.c import transform_flat_data_with_dict
@@ -18,14 +18,14 @@ class CategoricalConverter(Converter):
                  flat_arr: flat_arr_type) -> np.ndarray:
         flat_arr = np.asarray(flat_arr, np_float_type)
         if not self.info.need_transform:
-            return flat_arr
+            return flat_arr.astype(np_int_type)
         return transform_flat_data_with_dict(flat_arr, self._transform_dict)
 
     def _recover(self,
                  flat_arr: flat_arr_type) -> np.ndarray:
         flat_arr = np.asarray(flat_arr, np_float_type)
         if not self.info.need_transform:
-            return flat_arr
+            return flat_arr.astype(np_float_type)
         return transform_flat_data_with_dict(flat_arr, self._reverse_transform_dict)
 
 
