@@ -7,20 +7,27 @@ import numpy as np
 from typing import *
 from cftool.misc import timing_context, SavingMixin
 
-from .types import *
+from .misc import *
 from .recognizer import *
 from .converters import *
 from .processors import *
-from .utils import DataSplitter
 from ..base import DataBase
 from ..types import np_int_type
 
 
+class TabularSplit(NamedTuple):
+    split: "TabularData"
+    remained: "TabularData"
+    split_indices: np.ndarray
+    remained_indices: np.ndarray
+
+
 # TODO : Add outlier detection
 class TabularData(DataBase):
     def __init__(self,
                  *,
-                 task_type: TaskTypes = None,
+                 task_type: TaskTypes = TaskTypes.NONE,
+                 time_series_config: TimeSeriesConfig = None,
                  label_name: str = None,
                  string_label: Union[bool, None] = None,
                  numerical_label: Union[bool, None] = None,
@@ -37,16 +44,16 @@ class TabularData(DataBase):
                  numerical_threshold: float = None,
                  trigger_logging: bool = False,
                  verbose_level: int = 1):
-        if task_type is not None:
-            if task_type is TaskTypes.CLASSIFICATION:
-                if numerical_label:
-                    raise ValueError("numerical labels are invalid in CLASSIFICATION tasks")
-            else:
-                if string_label:
-                    raise ValueError("string labels are invalid in REGRESSION tasks")
-                if categorical_label:
-                    raise ValueError("categorical labels are invalid in REGRESSION tasks")
+        if task_type.is_clf:
+            if numerical_label:
+                raise ValueError("numerical labels are invalid in CLASSIFICATION tasks")
+        elif task_type.is_reg:
+            if string_label:
+                raise ValueError("string labels are invalid in REGRESSION tasks")
+            if categorical_label:
+                raise ValueError("categorical labels are invalid in REGRESSION tasks")
         self._task_type = task_type
+        self._time_series_config = time_series_config
         self.label_name = label_name
         self.string_label = string_label
         self.numerical_label = numerical_label
@@ -86,6 +93,32 @@ class TabularData(DataBase):
             return False
         return self.processed == other.processed
 
+    @property
+    def ts_config(self) -> Union[TimeSeriesConfig, None]:
+        if self._time_series_config is None:
+            return
+        id_name = self._time_series_config.id_column_name
+        time_name = self._time_series_config.time_column_name
+        id_idx = self._time_series_config.id_column_idx
+        time_idx = self._time_series_config.time_column_idx
+        if id_idx is None:
+            if id_name is None:
+                raise ValueError("either `id_column_name` or `id_column` should be provided")
+            for k, v in self.column_names.items():
+                if v == id_name:
+                    id_idx = k
+                    break
+        if time_idx is None:
+            if time_name is None:
+                raise ValueError("either `time_column_name` or `time_column` should be provided")
+            for k, v in self.column_names.items():
+                if v == time_name:
+                    time_idx = k
+                    break
+        id_column = self.raw.xT[id_idx]
+        time_column = self.raw.xT[time_idx]
+        return TimeSeriesConfig(id_name, time_name, id_idx, time_idx, id_column, time_column)
+
     @property
     def cache_excludes(self):
         return {"_recognizers", "_converters", "_processors"}
@@ -131,12 +164,15 @@ class TabularData(DataBase):
         return self._processed.x.shape[1]
 
     @property
-    def task_type(self) -> Union[TaskTypes, None]:
-        if self._task_type is not None:
+    def task_type(self) -> TaskTypes:
+        if not self._task_type.is_none:
             return self._task_type
         if self._recognizers[-1] is None:
-            return
-        self._task_type = TaskTypes.from_column_type(self._recognizers[-1].info.column_type)
+            return TaskTypes.NONE
+        self._task_type = TaskTypes.from_column_type(
+            self._recognizers[-1].info.column_type,
+            is_time_series=self.is_ts,
+        )
         return self._task_type
 
     @property
@@ -147,7 +183,7 @@ class TabularData(DataBase):
             self._column_names.setdefault(i, str(i))
         return self._column_names
 
-    def _get_prior_dict(self, attr: str) -> Dict[int, Union[bool, None]]:
+    def _get_prior_dict(self, attr: str, ts_value: Union[bool, None]) -> Dict[int, Union[bool, None]]:
         prior_dict_attr = f"{attr}_dict"
         prior_dict = getattr(self, prior_dict_attr, None)
         if prior_dict is None:
@@ -157,36 +193,54 @@ class TabularData(DataBase):
                 for i in range(self.raw_dim)
             }
             setattr(self, prior_dict_attr, prior_dict)
+        ts_config = self.ts_config
+        if ts_config is not None:
+            prior_dict[ts_config.id_column_idx] = ts_value
+            prior_dict[ts_config.time_column_idx] = ts_value
         return prior_dict
 
     @property
     def prior_valid_columns(self) -> Dict[int, Union[bool, None]]:
-        return self._get_prior_dict("_valid_columns")
+        return self._get_prior_dict("_valid_columns", True)
 
     @property
     def prior_string_columns(self) -> Dict[int, Union[bool, None]]:
-        return self._get_prior_dict("_string_columns")
+        return self._get_prior_dict("_string_columns", None)
 
     @property
     def prior_numerical_columns(self) -> Dict[int, Union[bool, None]]:
-        return self._get_prior_dict("_numerical_columns")
+        return self._get_prior_dict("_numerical_columns", False)
 
     @property
     def prior_categorical_columns(self) -> Dict[int, Union[bool, None]]:
-        return self._get_prior_dict("_categorical_columns")
+        return self._get_prior_dict("_categorical_columns", None)
 
     @property
     def is_clf(self) -> bool:
-        return self.task_type is TaskTypes.CLASSIFICATION
+        return self.task_type.is_clf
 
     @property
     def is_reg(self) -> bool:
-        return self.task_type is TaskTypes.REGRESSION
+        return self.task_type.is_reg
+
+    @property
+    def is_ts(self) -> bool:
+        return self.ts_config is not None
 
     @property
     def is_file(self) -> bool:
         return self._is_file
 
+    @property
+    def splitter(self) -> DataSplitter:
+        splitter = DataSplitter(time_series_config=self.ts_config, shuffle=False)
+        return splitter.fit(self.to_dataset())
+
+    @property
+    def ts_indices(self) -> Set[int]:
+        ts_config = self.ts_config
+        return set() if ts_config is None else {ts_config.id_column_idx, ts_config.time_column_idx}
+
     @property
     def num_classes(self) -> int:
         if self.is_reg:
@@ -199,11 +253,18 @@ class TabularData(DataBase):
 
     @staticmethod
     def _flatten(data: data_type) -> data_type:
-        if isinstance(data, list):
-            return sum(data, [])
-        return data.ravel()
+        if isinstance(data, np.ndarray):
+            return data.ravel()
+        flattened = []
+        for elem in data:
+            flattened.extend(elem)
+        return flattened
+
+    def _get_ts_sorting_indices(self):
+        self.ts_sorting_indices = np.hstack(self.splitter._time_indices_list_in_use)[::-1].copy()
 
     def _core_fit(self) -> "TabularData":
+        ts_indices = self.ts_indices
         with timing_context(self, "convert"):
             # convert features
             features = self._raw.xT
@@ -237,9 +298,10 @@ class TabularData(DataBase):
                     self.log_msg(recognizer.info.msg, self.warning_prefix, 2, logging.WARNING)
                     self.excludes.add(i)
                     continue
-                with timing_context(self, "fit converter"):
-                    converter = self._converters[i] = Converter.make_with(recognizer)
-                converted_features.append(converter.converted_input)
+                if i not in ts_indices:
+                    with timing_context(self, "fit converter"):
+                        converter = self._converters[i] = Converter.make_with(recognizer)
+                    converted_features.append(converter.converted_input)
             # convert labels
             if self._raw.y is None:
                 converted_labels = self._recognizers[-1] = self._converters[-1] = None
@@ -266,7 +328,7 @@ class TabularData(DataBase):
             previous_processors = []
             idx = 0
             while idx < self.raw_dim:
-                if idx in self.excludes:
+                if idx in self.excludes or idx in ts_indices:
                     idx += 1
                     continue
                 column_type = self._converters[idx].info.column_type
@@ -279,7 +341,9 @@ class TabularData(DataBase):
                 if method is None:
                     method = "identical"
                 elif method == "auto":
-                    if column_type is ColumnTypes.NUMERICAL:
+                    if idx in ts_indices:
+                        method = "identical"
+                    elif column_type is ColumnTypes.NUMERICAL:
                         method = self._default_numerical_process
                     else:
                         method = self._default_categorical_process
@@ -305,27 +369,30 @@ class TabularData(DataBase):
                     processor = self._processors[-1] = processor_dict[method]([]).fit(converted_labels)
                 with timing_context(self, "process with processor"):
                     processed_labels = processor.process(converted_labels)
-        if self.task_type is TaskTypes.CLASSIFICATION and converted_labels is not None:
+        if self.task_type.is_clf and converted_labels is not None:
             converted_labels = converted_labels.astype(np_int_type)
             processed_labels = processed_labels.astype(np_int_type)
         self._converted = DataTuple(converted_x, converted_labels)
         self._processed = DataTuple(np.hstack(processed_features), processed_labels)
         self._valid_columns = [col for col in range(self.raw_dim) if col not in self.excludes]
-        self._valid_columns_dict = None
+        self._valid_columns_dict = self.ts_sorting_indices = None
+        if self.is_ts:
+            self._get_ts_sorting_indices()
         return self
 
     def _read_from_file(self,
                         file_path: str,
                         *,
                         label_idx: int = None,
+                        contains_labels: bool = True,
                         has_column_names: bool = None,
                         quote_char: str = None,
                         delim: str = None) -> "TabularData":
         self._is_file = True
         self._label_idx, self._has_column_names = label_idx, has_column_names
         self._delim, self._quote_char = delim, quote_char
-        with timing_context(self, "_read_file"):
-            x, y = self.read_file(file_path)
+        with timing_context(self, "read_file"):
+            x, y = self.read_file(file_path, contains_labels=contains_labels)
         self._raw = DataTuple.with_transpose(x, y)
         return self._core_fit()
 
@@ -343,7 +410,7 @@ class TabularData(DataBase):
         else:
             converted_labels = self._converters[-1].convert(self._flatten(raw.y))
             transformed_labels = self._processors[-1].process(converted_labels.reshape([-1, 1]))
-        if self.task_type is TaskTypes.CLASSIFICATION:
+        if self.task_type.is_clf:
             if converted_labels is not None:
                 converted_labels = converted_labels.astype(np_int_type)
                 transformed_labels = transformed_labels.astype(np_int_type)
@@ -354,14 +421,16 @@ class TabularData(DataBase):
                    return_converted: bool) -> Union[DataTuple, Tuple[DataTuple, DataTuple]]:
         # transform features
         features = raw.xT
+        ts_indices = self.ts_indices
         converted_features = np.vstack([
             self._converters[i].convert(flat_arr)
-            for i, flat_arr in enumerate(features) if i not in self.excludes
+            for i, flat_arr in enumerate(features)
+            if i not in self.excludes and i not in ts_indices
         ])
         idx = 0
         processed = []
         while idx < self.raw_dim:
-            if idx in self.excludes:
+            if idx in self.excludes or idx in ts_indices:
                 idx += 1
                 continue
             processor = self._processors[idx]
@@ -413,6 +482,7 @@ class TabularData(DataBase):
             )
         else:
             raise NotImplementedError(f"file type '{ext}' not recognized")
+        self._delim, self._quote_char = delim, quote_char
         with open(file_path, "r") as f:
             first_row = None
             if has_column_names:
@@ -448,7 +518,7 @@ class TabularData(DataBase):
                     assert len(first_row) == len(line), "num_features are not identical"
                 data.append(line)
         if not contains_labels:
-            if len(data[0]) == len(self._raw.x[0]) + 1:
+            if self._raw is not None and len(data[0]) == len(self._raw.x[0]) + 1:
                 raise ValueError("file contains labels but 'contains_labels=False' passed in")
             return data, None
 
@@ -472,9 +542,11 @@ class TabularData(DataBase):
     def read(self,
              x: Union[str, data_type],
              y: Union[int, data_type] = None,
+             *,
+             contains_labels: bool = True,
              **kwargs) -> "TabularData":
         if isinstance(x, str):
-            self._read_from_file(x, label_idx=y, **kwargs)
+            self._read_from_file(x, label_idx=y, contains_labels=contains_labels, **kwargs)
         else:
             if isinstance(y, int):
                 y = None
@@ -483,11 +555,21 @@ class TabularData(DataBase):
         return self
 
     def split(self,
-              n: Union[int, float]) -> Tuple["TabularData", "TabularData"]:
-        splitter = DataSplitter(shuffle=False).fit(self.to_dataset())
-        split = splitter.split(n)
-        split_indices = split.corresponding_indices
-        remained_indices = split.remaining_indices
+              n: Union[int, float],
+              *,
+              order: str = "auto") -> TabularSplit:
+        if order == "auto":
+            split = self.splitter.split(n)
+            split_indices = split.corresponding_indices
+            remained_indices = split.remaining_indices
+        else:
+            if order not in {"bottom_up", "top_down"}:
+                raise NotImplementedError(f"`order` should be either 'bottom_up' or 'top_down', {order} found")
+            base_indices = np.arange(self._raw.x.shape[0])
+            if order == "bottom_up":
+                split_indices, remained_indices = base_indices[-n:], base_indices[:-n]
+            else:
+                split_indices, remained_indices = base_indices[:n], base_indices[n:]
         raw, converted, processed = self._raw, self._converted, self._processed
         p1, p2 = map(copy.copy, [self] * 2)
         p1._raw, p1._converted, p1._processed = map(
@@ -498,7 +580,9 @@ class TabularData(DataBase):
             DataTuple.split_with,
             [raw, converted, processed], [remained_indices] * 3
         )
-        return p1, p2
+        p1.ts_sorting_indices = np.arange(len(p1))
+        p2.ts_sorting_indices = np.arange(len(p2))
+        return TabularSplit(p1, p2, split_indices, remained_indices)
 
     def copy_to(self,
                 x: Union[str, data_type],
@@ -508,14 +592,17 @@ class TabularData(DataBase):
         copied = copy.copy(self)
         raw = copied._raw = self._get_raw(x, y, contains_labels=contains_labels)
         copied._converted, copied._processed = self._transform(raw, True)
+        if copied.is_ts:
+            copied._get_ts_sorting_indices()
         return copied
 
     def transform(self,
                   x: Union[str, data_type],
                   y: data_type = None,
                   *,
+                  contains_labels: bool = True,
                   return_converted: bool = False) -> Union[DataTuple, Tuple[DataTuple, DataTuple]]:
-        raw = self._get_raw(x, y)
+        raw = self._get_raw(x, y, contains_labels=contains_labels)
         return self._transform(raw, return_converted)
 
     def transform_labels(self,
@@ -555,9 +642,20 @@ class TabularData(DataBase):
     @classmethod
     def from_dataset(cls,
                      dataset: TabularDataset,
-                     **kwargs):
+                     **kwargs) -> "TabularData":
         task_type = kwargs.pop("task_type", dataset.task_type)
         return cls(task_type=task_type, **kwargs).read(*dataset.xy)
 
+    @classmethod
+    def simple(cls,
+               task_type: TaskTypes,
+               **kwargs) -> "TabularData":
+        return cls(
+            default_numerical_process="identical",
+            default_categorical_process="identical",
+            task_type=task_type, verbose_level=0,
+            **kwargs
+        )
+
 
 __all__ = ["TabularData"]
